import numpy as np
from jenn.tests.test_problems import rosenbrock
from jenn._utils import grad_check


def test_grad_check():
    """
    Test that method to check gradient accuracy works, using the
    banana Rosenbrock function as a test example
    """
    x0 = [np.array([1.25, -1.75]).reshape((2, 1))]
    f = lambda x: rosenbrock(x)[0]
    dfdx = lambda x: rosenbrock(x)[1]
    assert grad_check(x0, f, dfdx)


"""
J A C O B I A N - E N H A N C E D   N E U R A L   N E T W O R K S  (J E N N)

Author: Steven H. Berguin <stevenberguin@gmail.com>

This package is distributed under the MIT license.
"""

import numpy as np
import math
from typing import List

from importlib.util import find_spec

if find_spec("matplotlib"):
    import matplotlib.pyplot as plt
    import matplotlib.gridspec as gridspec
    MATPLOTLIB_INSTALLED = True
else:
    MATPLOTLIB_INSTALLED = False

EPS = np.finfo(float).eps  # small number to avoid division by zero


def mini_batches(X: np.ndarray, batch_size: int,
                 shuffle: bool = True, random_state: int = None) -> list:
    """
    Create randomized mini-batches by returning a list of tuples, where
    each tuple contains the indices of the training data points associated with
    that mini-batch

    Parameters
    ----------
        X: np.ndarray
            input features of the training data
            shape = (n_x, m) where m = num of examples and n_x = num of inputs

        batch_size: int
            mini batch size (if None, then batch_size = m)
            
        shuffle: bool 
            Shuffle data points
            Default = True 
            
        random_state: int 
            Random seed (set to make runs repeatable)
            Default = None  

    Returns
    -------
        mini_batches: list
            List of mini-batch indices to use for slicing data, where the index
            is in the interval [1, m]
    """

    np.random.seed(random_state)

    batches = []
    m = X.shape[1]
    if not batch_size:
        batch_size = m
    batch_size = min(batch_size, m)

    # Step 1: Shuffle the indices
    if shuffle:
        indices = list(np.random.permutation(m))
    else:
        indices = np.arange(m)

    # Step 2: Partition (shuffled_X, shuffled_Y). Minus the end case.
    num_complete_minibatches = int(math.floor(m / batch_size))
    k = 0
    for _ in range(num_complete_minibatches):
        mini_batch = indices[k * batch_size:(k + 1) * batch_size]
        if mini_batch:
            batches.append(tuple(mini_batch))
        k += 1

    # Handling the end case (last mini-batch < mini_batch_size)
    if m % batch_size != 0:
        mini_batch = indices[(k + 1) * batch_size:]
        if mini_batch:
            batches.append(tuple(mini_batch))

    return batches


def finite_diff(params: List[np.ndarray], f: callable, dx: float = 1e-6):
    """
    Compute gradient using central difference

    Parameters
    ----------
    params: List[np.ndarray]
        List of neural net parameters at which to evaluate gradient

    f: callable
        Function handle to use for finite difference

    dx: float
        Finite difference step

    Returns
    -------
    dy_dx: List[np.ndarray]
        Derivative of f w.r.t. x
    """
    is_list = True
    if type(params) is np.ndarray:
        params = [params]
        is_list = False
    grads = list()
    for k, x in enumerate(params):
        n, p = x.shape
        dy = np.zeros((n, p))
        for i in range(0, n):
            for j in range(0, p):
                # Forward step
                x[i, j] += dx
                if is_list:
                    y_fwd = f(params)
                else:
                   y_fwd = f(params[0])
                x[i, j] -= dx

                # Backward step
                x[i, j] -= dx
                if is_list:
                    y_bwd = f(params)
                else:
                    y_bwd = f(params[0])
                x[i, j] += dx

                # Central difference
                dy[i, j] = np.divide(y_fwd - y_bwd, 2 * dx)

        grads.append(dy)

    return grads


def grad_check(x: List[np.ndarray], f: callable, dfdx: callable,
               dx: float = 1e-6, tol: float = 1e-6,
               verbose: bool = True) -> bool:
    """
    Compare analytical gradient against finite difference

    Parameters
    ----------
    x: List[np.ndarray]
        Point at which to evaluate gradient

    f: callable
        Function handle to use for finite difference

    dx: float
        Finite difference step

    tol: float
        Tolerance below which agreement is considered acceptable
        Default = 1e-6

    verbose: bool
        Print output to standard out
        Default = True

    Returns
    -------
    success: bool 
        Returns True iff finite difference and analytical grads agree 
    """
    success = True
    dydx = dfdx(x)
    dydx_FD = finite_diff(x, f, dx)
    for i in range(len(x)):
        numerator = np.linalg.norm(dydx[i].squeeze() - dydx_FD[i].squeeze())
        denominator = np.linalg.norm(dydx[i].squeeze()) + np.linalg.norm(
            dydx_FD[i].squeeze())
        difference = numerator / (denominator + EPS)
        if difference > tol or numerator > tol:
            success = False
        if verbose:
            if not success:
                print(f"The gradient w.r.t. x[{i}] is wrong")
            else:
                print(f"The gradient w.r.t. x[{i}] is correct")
            print(f"Finite dif: grad[{i}] = {dydx_FD[i].squeeze()}")
            print(f"Analytical: grad[{i}] = {dydx[i].squeeze()}")
            print()
    return success


def rsquare(y_pred, y_true):
    epsilon = 1e-12  # small number to avoid division by zero
    y_bar = np.mean(y_true)
    SSE = np.sum(np.square(y_pred - y_true))
    SSTO = np.sum(np.square(y_true - y_bar) + epsilon)
    R2 = 1 - SSE / SSTO
    return R2


def goodness_of_fit(y_pred: np.ndarray, y_true: np.ndarray,
                    title: str = None, legend: str = None,
                    show_error: bool = True):
    """
    Plot actual by predicted and histogram of prediction error

    Parameters
    ----------

    y_pred: np.ndarray
        Predicted values
        shape = (m,) where m = no. examples

    y_true: np.ndarray
        True values
        shape = (m,) where m = no. examples

    title: str
        Title to be displayed on figure
        Default = None => 'Goodness of Fit'

    legend: str
        Labeled to be displayed in legend for data
        Default = None => 'data samples'
    """
    # Prepare to plot
    if not MATPLOTLIB_INSTALLED:
        raise ImportError("Matplotlib must be installed.")

    if not title:
        title = 'Goodness of Fit'

    if not legend:
        legend = 'data samples'

    r_squared = rsquare(y_pred, y_true)
    std_error = np.std(y_pred - y_true)
    avg_error = np.mean(y_pred - y_true)

    # Reference Line
    y = np.linspace(np.min(y_true), np.max(y_true), 100)

    ncols = 1 + show_error
    fig = plt.figure(figsize=(6 * ncols, 6))
    fig.suptitle(title, fontsize=16)
    spec = gridspec.GridSpec(ncols=ncols, nrows=1, wspace=0.25)

    # Plot
    ax1 = fig.add_subplot(spec[0, 0])
    ax1.plot(y, y, 'r-')
    ax1.scatter(y_true, y_pred, s=100, c='k', marker=".")
    ax1.legend(["perfect fit line", legend])
    ax1.set_xlabel("actual")
    ax1.set_ylabel("predicted")
    ax1.set_title(f"RSquare = {r_squared:.2f}")
    plt.grid(True)

    if not show_error:
        return fig

    ax2 = fig.add_subplot(spec[0, 1])
    error = (y_pred - y_true)
    weights = np.ones(error.shape) / y_pred.size
    ax2.hist(error, weights=weights, facecolor='g', alpha=0.75, bins=20)
    ax2.set_xlabel('Absolute Prediction Error')
    ax2.set_ylabel('Probability')
    ax2.set_title(f'$\mu$ = {avg_error:.2f}, $\sigma$ = {std_error:.2f}')
    plt.grid(True)
    plt.show()

    return fig


class DataConverter:
    """ Helper class to normalize and denormalize data """

    def X(self, X_norm: np.ndarray):
        """
        Return denormalized inputs: X = X_norm * sigma_x + mu_x

        Parameters
        ----------
        X_norm: np.ndarray
            Normalized input, shape (n_x, m) where n_x = no. of inputs
                                                   m = no. of training examples
        """
        return X_norm * self._sigma_x + self._mu_x

    def Y(self, Y_norm: np.ndarray):
        """
        Return denormalized outputs: Y = Y_norm * sigma_y + mu_y

        Parameters
        ----------
        Y_norm: np.ndarray
            Normalized output, shape (n_y, m) where n_y = no. of outputs
                                                   m = no. of training examples
        """
        return Y_norm * self._sigma_y + self._mu_y

    def J(self, J_norm: np.ndarray):
        """
        Return denormalized Jacobian: J_norm = J_norm * sigma_y / sigma_x

        Parameters
        ----------
        J_norm: np.ndarray
            Normalized Jacobian, shape (n_y, n_x, m) where
                                                   n_y = no. of outputs
                                                   n_x = no. of inputs
                                                   m = no. of training examples
        """
        n_y, n_x, _ = J_norm.shape
        sigma_x = self._sigma_x.T.reshape((1, n_x, 1))
        sigma_y = self._sigma_y.reshape((n_y, 1, 1))
        return J_norm * sigma_y / sigma_x

    def X_norm(self, X: np.ndarray):
        """
        Return normalized inputs: X_norm = (X - mu_x) / sigma_y

        Parameters
        ----------
        X: np.ndarray
            Input, shape (n_x, m) where n_x = no. of inputs
                                        m = no. of training examples
        """
        return (X - self._mu_x) / (self._sigma_x + 1e-12)

    def Y_norm(self, Y: np.ndarray):
        """
        Return normalized outputs: Y_norm = (Y - mu_y) / sigma_y

        Parameters
        ----------
        Y: np.ndarray
            Output, shape (n_y, m) where n_y = no. of outputs
                                         m = no. of training examples
        """
        return (Y - self._mu_y) / (self._sigma_y + 1e-12)

    def J_norm(self, J: np.ndarray):
        """
        Return normalized Jacobian: J_norm = J * sigma_x / sigma_y

        Parameters
        ----------
        J: np.ndarray
            Jacobian, shape (n_y, n_x, m) where    n_y = no. of outputs
                                                   n_x = no. of inputs
                                                   m = no. of training examples
        """
        if J is not None:
            n_y, n_x, _ = J.shape
            sigma_x = self._sigma_x.T.reshape((1, n_x, 1))
            sigma_y = self._sigma_y.reshape((n_y, 1, 1))
            return J * sigma_x / sigma_y

    def __init__(self,
                 mu_x: np.ndarray, sigma_x: np.ndarray,
                 mu_y: np.ndarray, sigma_y: np.ndarray):
        """
        X_norm = (X - mu_x) / sigma_y
        Y_norm = (Y - mu_y) / sigma_y

        Parameters
        ----------
        mu_x: np.ndarray
            Array of shape (n_x, 1) where n_x = no. of inputs

        mu_y: np.ndarray
            Array of shape (n_y, 1) where n_y = no. of outputs

        sigma_x: np.ndarray
            Array of shape (n_x, 1) where n_x = no. of inputs

        sigma_y: np.ndarray
            Array of shape (n_y, 1) where n_y = no. of outputs
        """
        self._mu_x = mu_x
        self._mu_y = mu_y
        self._sigma_x = sigma_x
        self._sigma_y = sigma_y


def scale_factor(array: np.ndarray) -> tuple:
    """
    Find mu and sigma such that X_norm = (X - mu) / sigma

    Parameters
    ----------
    array: np.ndarray
        Input features of shape (n, m) where n = dimensionality
                                             m = number of examples
    """
    mu = np.mean(array, axis=1).reshape((-1, 1))
    sigma = np.std(array, axis=1).reshape((-1, 1))
    return mu, sigma




"""
J A C O B I A N - E N H A N C E D   N E U R A L   N E T W O R K S  (J E N N)

Author: Steven H. Berguin <stevenberguin@gmail.com>

This package is distributed under the MIT license.
"""

from ._activation import ACTIVATIONS

import numpy as np
from typing import List


def linear_activation_forward(A_prev: np.ndarray, W: np.ndarray, b: np.ndarray,
                              activation: str, store_cache: bool = True):
    """
    Implements forward propagation for one layer

    Parameters
    ----------
    A_prev: np.ndarray
        activations from previous layer
        shape = (n[l-1], 1) where n[l-1] = no. nodes in previous layer

    W: np.ndarray
        weights associated with current layer l
        shape = (n_l, n[l-1]) where n_l = no. nodes in current layer

    b: np.ndarray
        intercepts associated with current layer
        shape = numpy array of size (n_l, 1)

    activation: str
        layer activation type = {'identity', 'tanh', 'logistic', 'relu'}

    store_cache: bool
        True = do not compute cache (applies to trained model)
        False = compute cache (applies to untrained model) = default

    Return
    ------
    A: np.ndarray
        post-activation values for current layer
        shape = numpy array of size (n_l, 1)

    cache: tuple
        computed parameters that can be re-used for back-prop
        to avoid duplicate computations. Tuple contains:

            A_prev = numpy array of shape (n[l-1], m)
                     contains previous layer post-activation values where:
                        n[l-1] = no. nodes in previous layer
                        m = no. of training examples

            Z = numpy array of shape (n[l], m)
                contains linear forward values
                where n_l = number of nodes in current layer

            W = numpy array of shape (n[l], n[l-1]) containing weights
                of the current layer

            b = numpy array of shape (n[l], 1) containing biases
                of the current layer
    """
    g = ACTIVATIONS[activation]
    Z = np.dot(W, A_prev) + b
    A = g.evaluate(Z)
    if store_cache:
        cache = (A_prev, Z, W, b, activation)
    else:
        cache = None
    return A, cache


def L_model_forward(X: np.ndarray, W: List[np.ndarray], b: List[np.ndarray],
                    activations: List[str], store_cache: bool = True):
    """
    Implements forward propagation for the entire neural network

    Parameters
    ----------
    X: np.ndarray
        input layer activations
        shape = (n_x, m) where n_x = number of inputs
                               m = number of training examples

    W: List[np.ndarray]
        List containing weights associated with each layer
        shape = (n_i, n[i-1]) where n_i = no. nodes in i^th layer

    b: List[np.ndarray]
        List containing intercepts associated with each layer
        shape = (n_i, 1) where n_i = no. nodes in i^th layer

    activations: List[str]
        List containing activation types for each layer (hidden + output layer)
        Allowed values = {'identity', 'tanh', 'logistic', 'relu'}

    store_cache: bool
        True = do not compute cache (applies to trained model)
        False = compute cache (applies to untrained model) = default

    Return
    ------
    AL: np.ndarray
        post-activation value of output layer (layer L)
        shape = (n_y, m) where n_y = number of outputs
                               m = number of training examples

    caches: List[tuple]
        Only returned if fast = False
        List containing the cache for each layer where:
            cache = computed parameters that can be re-used for back-prop
                    to avoid duplicate computations. Tuple contains:

                    A_prev = numpy array of shape (n[l-1], m)
                             contains previous layer post-activation values:
                                n[l-1] = no. nodes in previous layer
                                m = no. of training examples

                    Z = numpy array of shape (n[l], m)
                        contains linear forward values where
                        n_l = number of nodes in current layer

                    W = numpy array of shape (n[l], n[l-1])
                        containing weights of the current layer

                    b = numpy array of shape (n[l], 1)
                        containing biases of the current layer
    """
    caches = []
    A = X
    L = len(activations)  # num layers in network (doesn't include input layer)
    for l in range(L):
        A, cache = linear_activation_forward(A, W[l], b[l],
                                             activations[l], store_cache)
        if store_cache:
           caches.append(cache)
    if store_cache:
        return A, caches
    return A


# def linear_grad_forward(A_prev: np.ndarray, A_prime_j_prev: np.ndarray,
#                         W: np.ndarray, b: np.ndarray, j: int,
#                         activation: str, store_cache: bool = True):
#     """
#     Implements forward propagation of gradient for one layer
#
#     Parameters
#     ----------
#     A_prev: np.ndarray
#         activations from previous layer
#         shape = (n[l-1], 1) where n[l-1] = no. nodes in previous layer
#
#     A_prime_j_prev: np.ndarray
#         previous layer partials of activations w.r.t. input layer X
#         shape = (n[l-1], 1) where n[l-1] = no. nodes in previous layer
#
#     W: np.ndarray
#         weights associated with current layer l
#         shape = (n_l, n[l-1]) where n_l = no. nodes in current layer
#
#     b: np.ndarray
#         intercepts associated with current layer
#         shape = numpy array of size (n_l, 1)
#
#     j: int
#         The j^th input w.r.t. which this partial is computed, i.e.  X[j, :]
#
#     activation: str
#         layer activation type = {'identity', 'tanh', 'logistic', 'relu'}
#
#     store_cache: bool
#         True = do not compute cache (applies to trained model)
#         False = compute cache (applies to untrained model) = default
#
#     Return
#     ------
#     A_prime_j: np.ndarray
#         partial of post-activation values w.r.t. X[j] for current layer
#         shape = numpy array of size (n_l, 1)
#
#     J_cache: tuple
#         computed parameters that can be re-used for back-prop
#         to avoid duplicate computations. Tuple contains:
#
#             j = index of input with respect to which partial is computed
#
#             Z_prime_j = partial of Z with respect to X[j]
#                         shape = (n[l-1], m)
#                         where n[l-1] = number of nodes in previous layer
#                               m = number of training examples
#
#             A_prime_j_prev = partial of previous layer A w.r.t. X[j]
#                              shape = (n[l-1], m)
#                              where n[l-1] = number nodes in previous layer
#                                    m = no. of training examples
#     """
#     g = ACTIVATIONS[activation]
#     Z = np.dot(W, A_prev) + b
#     G_prime = g.first_derivative(Z)
#     A_prime_j = G_prime * np.dot(W, A_prime_j_prev)
#     J_cache = None
#     if store_cache:
#         Z_prime_j = np.dot(W, A_prime_j_prev)
#         J_cache = (j, Z_prime_j, A_prime_j_prev)
#     return A_prime_j, J_cache
#
#
# def L_grads_forward(X: np.ndarray, W: List[np.ndarray], b: List[np.ndarray],
#                     activations: List[str], store_cache: bool = True):
#     """
#     Implements forward propagation of gradient for the entire neural network
#
#     Parameters
#     ----------
#     X: np.ndarray
#         input layer activations
#         shape = (n_x, m) where n_x = number of inputs
#                                m = number of training examples
#
#     W: List[np.ndarray]
#         List containing weights associated with each layer
#         shape = (n_i, n[i-1]) where n_i = number nodes in i^th layer
#
#     b: List[np.ndarray]
#         List containing intercepts associated with each layer
#         shape = (n_i, 1) where n_i = no. nodes in i^th layer
#
#     activations: List[str]
#         List of activation types for each layer (hidden + output layer)
#         Allowed values = {'identity', 'tanh', 'logistic', 'relu'}
#
#     store_cache: bool
#         True = do not compute cache (applies to trained model)
#         False = compute cache (applies to untrained model) = default
#
#     Return
#     ------
#     AL: np.ndarray
#         post-activation value of output layer (layer L)
#         shape = (n_y, m) where n_y = number of outputs
#                                m = number of training examples
#
#     JL: np.ndarray
#         Jacobian of w.r.t. X (i.e. partials of outputs w.r.t inputs)
#         shape = (n_y, n_x, m) where n_y = no. of outputs
#
#     J_caches: List[tuple]
#         Only returned if fast = False
#         List containing the Jacobian cache for each layer where:
#         J_cache = [..., (j, Z_prime_j, A_prime_j, G_prime, G_prime_prime), ...]
#                        <------------------ input j ------------------>
#         where
#             j -- input variable number (i.e. X1, X2, ...) associated with cache
#                       >> an integer representing the associated input
#                          variables (X1, X2, ..., Xj, ...)
#             Z_prime_j -- derivative of Z w.r.t. X_j: Z'_j = d(Z_j)/dX_j
#                       >> a numpy array of shape (n_l, m) where n_l is the
#                          number of nodes in layer l
#             A_prime_j -- derivative of the activation w.r.t. X_j
#                             A'_j = d(A_j)/dX_j
#                       >> a numpy array of shape (n_l, m) where n_l is the
#                          number nodes in layer l
#     """
#     J_caches = []
#
#     # Dimensions
#     L = len(activations)  # number of layers in network
#     n_y = W[-1].shape[0]  # number of outputs
#     try:
#         n_x, m = X.shape  # number of inputs, number of examples
#     except ValueError:
#         n_x = X.size
#         m = 1
#         X = X.reshape(n_x, m)
#
#     # Initialize Jacobian for layer 0 (one example)
#     I = np.eye(n_x, dtype=float)
#
#     # Initialize Jacobian for layer 0 (all m examples)
#     J0 = np.repeat(I.reshape((n_x, n_x, 1)), m, axis=2)
#
#     # Initialize Jacobian for last layer
#     JL = np.zeros((n_y, n_x, m))
#
#     # Initialize caches
#     if store_cache:
#         for l in range(L):
#             J_caches.append([])
#
#     # Loop over partials
#     for j in range(0, n_x):
#
#         # Initialize (first layer)
#         A = np.copy(X).reshape(n_x, m)
#         A_prime_j = J0[:, j, :]
#
#         # Loop over layers
#         for l in range(L):
#
#             # Previous layer
#             A_prev = A
#             A_prime_j_prev = A_prime_j
#
#             # Current layer
#             A, _ = linear_activation_forward(A_prev, W[l], b[l],
#                                              activations[l],
#                                              store_cache)
#             A_prime_j, J_cache = linear_grad_forward(A_prev, A_prime_j_prev,
#                                                      W[l], b[l], j,
#                                                      activations[l],
#                                                      store_cache)
#
#             # Store cache
#             if store_cache:
#                 J_caches[l].append(J_cache)
#
#         # Last layer partials
#         JL[:, j, :] = A_prime_j
#
#     if store_cache:
#         return JL, J_caches
#     return JL


def L_grads_forward(X: np.ndarray, W: List[np.ndarray], b: List[np.ndarray],
                    activations: List[str], store_cache: bool = True):
    """
    Implements forward propagation of gradient for the entire neural network

    Parameters
    ----------
    X: np.ndarray
        input layer activations
        shape = (n_x, m) where n_x = number of inputs
                               m = number of training examples

    W: List[np.ndarray]
        List containing weights associated with each layer
        shape = (n_i, n[i-1]) where n_i = number nodes in i^th layer

    b: List[np.ndarray]
        List containing intercepts associated with each layer
        shape = (n_i, 1) where n_i = no. nodes in i^th layer

    activations: List[str]
        List of activation types for each layer (hidden + output layer)
        Allowed values = {'identity', 'tanh', 'logistic', 'relu'}

    store_cache: bool
        True = do not compute cache (applies to trained model)
        False = compute cache (applies to untrained model) = default

    Return
    ------
    AL: np.ndarray
        post-activation value of output layer (layer L)
        shape = (n_y, m) where n_y = number of outputs
                               m = number of training examples

    JL: np.ndarray
        Jacobian of w.r.t. X (i.e. partials of outputs w.r.t inputs)
        shape = (n_y, n_x, m) where n_y = no. of outputs

    J_caches: List[tuple]
        Only returned if fast = False
        List containing the Jacobian cache for each layer where:
        J_cache = [..., (j, Z_prime_j, A_prime_j, G_prime, G_prime_prime), ...]
                       <------------------ input j ------------------>
        where
            j -- input variable number (i.e. X1, X2, ...) associated with cache
                      >> an integer representing the associated input
                         variables (X1, X2, ..., Xj, ...)
            Z_prime_j -- derivative of Z w.r.t. X_j: Z'_j = d(Z_j)/dX_j
                      >> a numpy array of shape (n_l, m) where n_l is the
                         number of nodes in layer l
            A_prime_j -- derivative of the activation w.r.t. X_j
                            A'_j = d(A_j)/dX_j
                      >> a numpy array of shape (n_l, m) where n_l is the
                         number nodes in layer l
    """
    J_caches = []

    # Dimensions
    L = len(activations)  # number of layers in network
    n_y = W[-1].shape[0]  # number of outputs
    try:
        n_x, m = X.shape  # number of inputs, number of examples
    except ValueError:
        n_x = X.size
        m = 1
        X = X.reshape(n_x, m)

    # Initialize Jacobian for layer 0 (one example)
    I = np.eye(n_x, dtype=float)

    # Initialize Jacobian for layer 0 (all m examples)
    J0 = np.repeat(I.reshape((n_x, n_x, 1)), m, axis=2)

    # Initialize Jacobian for last layer
    JL = np.zeros((n_y, n_x, m))

    # Initialize caches
    if store_cache:
        for l in range(0, L):
            J_caches.append([])

    # Loop over partials
    for j in range(n_x):

        # Initialize (first layer)
        A = np.copy(X).reshape(n_x, m)
        A_prime_j = J0[:, j, :]

        # Loop over layers
        for l in range(L):
            # Previous layer
            A_prev = A
            A_prime_j_prev = A_prime_j

            # Get parameters for this layer
            key = activations[l]
            g = ACTIVATIONS[key]

            # Linear
            Z = np.dot(W[l], A_prev) + b[l]

            # The following is not needed here, but it is needed later, during backprop.
            # We will thus compute it here and store it as a cache for later use.
            Z_prime_j = np.dot(W[l], A_prime_j_prev)

            # Activation
            A = g.evaluate(Z)
            G_prime = g.first_derivative(Z)

            # Current layer output gradient
            A_prime_j = G_prime * np.dot(W[l], A_prime_j_prev)

            # Store cache
            if store_cache:
                J_caches[l].append((j, Z_prime_j, A_prime_j_prev))

        # Store partial
        JL[:, j, :] = A_prime_j

    if store_cache:
        return JL, J_caches
    return JL


"""
J A C O B I A N - E N H A N C E D   N E U R A L   N E T W O R K S  (J E N N)

Author: Steven H. Berguin <stevenberguin@gmail.com>

This package is distributed under the MIT license.
"""
from importlib.util import find_spec

if find_spec("matplotlib"):
    from matplotlib import pyplot as plt
    MATPLOTLIB_INSTALLED = True
else:
    MATPLOTLIB_INSTALLED = False

import numpy as np

tensor = np.ndarray


class Singleton:

    _instance = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = object.__new__(cls)
        return cls._instance


class Activation(Singleton):

    def evaluate(self, z):
        """
        Evaluate activation function

        :param z: a scalar or numpy array of any size
        :return: activation value at z
        """
        raise NotImplementedError

    def first_derivative(self, z):
        """
        Evaluate gradient of activation function

        :param z: a scalar or numpy array of any size
        :return: gradient at z
        """
        raise NotImplementedError

    def second_derivative(self, z):
        """
        Evaluate second derivative of activation function

        :param z: a scalar or numpy array of any size
        :return: second derivative at z
        """
        raise NotImplementedError


class Sigmoid(Activation):

    def evaluate(self, z):
        a = 1. / (1. + np.exp(-z))
        return a

    def first_derivative(self, z):
        a = self.evaluate(z)
        da = a * (1. - a)
        return da

    def second_derivative(self, z):
        a = self.evaluate(z)
        da = self.first_derivative(z)
        dda = da * (1 - 2 * a)
        return dda


class Tanh(Activation):

    def evaluate(self, z):
        numerator = np.exp(z) - np.exp(-z)
        denominator = np.exp(z) + np.exp(-z)
        a = np.divide(numerator, denominator)
        return a

    def first_derivative(self, z):
        a = self.evaluate(z)
        da = 1 - np.square(a)
        return da

    def second_derivative(self, z):
        a = self.evaluate(z)
        da = self.first_derivative(z)
        dda = -2 * a * da
        return dda


class Relu(Activation):

    def evaluate(self, z):
        a = (z > 0) * z
        return a

    def first_derivative(self, z):
        da = 1.0 * (z > 0)
        return da

    def second_derivative(self, z):
        dda = 0.0
        return dda


class Linear(Activation):

    def evaluate(self, z):
        return z

    def first_derivative(self, z):
        return np.ones(z.shape)

    def second_derivative(self, z):
        return np.zeros(z.shape)


ACTIVATIONS = {'identity': Linear(),
               'tanh': Tanh(),
               'logistic': Sigmoid(),
               'relu': Relu()}


if __name__ == "__main__":
    if not MATPLOTLIB_INSTALLED:
        raise ImportError("Matplotlib must be installed.")

    x = np.linspace(-10, 10, 100)
    activations = {'tanh': Tanh(), 'sigmoid': Sigmoid(), 'relu': Relu()}
    for name, activation in activations.items():
        plt.plot(x, activation.evaluate(x))
        plt.title(name)
        plt.show()





"""
J A C O B I A N - E N H A N C E D   N E U R A L   N E T W O R K S  (J E N N)

Author: Steven H. Berguin <stevenberguin@gmail.com>

This package is distributed under the MIT license.
"""

from abc import ABC, abstractmethod
import numpy as np
from typing import List

EPS = np.finfo(float).eps  # small number to avoid division by zero


class Update(ABC):

    @abstractmethod
    def _update(self, params: List[np.ndarray], grads: List[np.ndarray],
                alpha: float) -> List[np.ndarray]:
        raise NotImplementedError

    def __call__(self, params: List[np.ndarray], grads: List[np.ndarray],
                 alpha: float) -> List[np.ndarray]:
        return self._update(params, grads, alpha)


class GD(Update):

    def _update(self, params: List[np.ndarray], grads: List[np.ndarray],
                alpha: float):
        new_params = []
        for k, x in enumerate(params):
            new_params.append(x - alpha * grads[k])
        return new_params


class ADAM(Update):

    def __init__(self, beta_1: float = 0.9, beta_2: float = 0.99):
        """
        Initialize ADAM updating 
        
        Parameters
        ----------
        beta_1: float
            Exponential decay rate for estimates of first moment vector
            in adam, should be in [0, 1). Only used when solver=’adam’
            Default = 0.9

        beta_2: float
            Exponential decay rate for estimates of second moment vector
            in adam, should be in [0, 1). Only used when solver=’adam’
            Default = 0.99
        """
        self.beta_1 = beta_1
        self.beta_2 = beta_2

        self._v = None
        self._s = None
        self._t = 0
        self._grads = None

    def _update(self, params: List[np.ndarray], grads: List[np.ndarray],
                alpha: float) -> List[np.ndarray]:
        """
        Take a single step in direction of improvement according to ADAM

        Parameters
        ----------
        params: List[np.ndarray]
            List of parameters to be updated

        grads: List[np.ndarray]
            Gradient of objective function w.r.t. each parameter

        alpha: float
            Learning rate

        Returns
        -------
        new_params: List[np.ndarray]
            List of updated parameters
        """

        beta_1 = self.beta_1
        beta_2 = self.beta_2

        v = self._v
        s = self._s
        t = self._t

        if not v:
            v = [np.zeros(param.shape) for param in params]
        if not s:
            s = [np.zeros(param.shape) for param in params]

        a = [id(x) for x in grads]
        b = []
        if self._grads:
            b = [id(x) for x in self._grads]

        if a != b:
            self._grads = grads
            t += 1  # only update for new search directions

        new_params = []
        for k, x in enumerate(params):
            v[k] = beta_1 * v[k] + (1. - beta_1) * grads[k]
            s[k] = beta_2 * s[k] + (1. - beta_2) * np.square(grads[k])
            v_corrected = v[k] / (1. - beta_1 ** t)
            s_corrected = s[k] / (1. - beta_2 ** t) + EPS
            new_params.append(x - alpha * v_corrected / np.sqrt(s_corrected))

        self._v = v
        self._s = s
        self._t = t

        return new_params


class LineSearch(ABC):

    def __init__(self, update: Update):
        self.update = update

    @abstractmethod
    def search(self, params: List[np.ndarray], grads: List[np.ndarray],
               cost: callable, learning_rate: float):
        raise NotImplementedError


class Backtracking(LineSearch):

    def __init__(self, update: Update, tau: float = 0.5,
                 tol: float = 1e-6, max_count: int = 1_000):
        super().__init__(update)
        self.tau = tau
        self.tol = tol
        self.max_count = max_count

    def search(self,
               params: List[np.ndarray], grads: List[np.ndarray],
               cost: callable, learning_rate: float = 0.1):
        """
        Take multiple update steps along search direction determined by update

        Parameters
        ----------
        params: List[np.ndarray]
            List of parameters to be updated

        grads: List[np.ndarray]
            Gradient of objective function w.r.t. each parameter

        cost: callable
            Objective function y = f(x) where x = params

        learning_rate: float
            The maximum allowed step size
        """
        tau = self.tau
        tol = self.tol
        x0 = self.update(params, grads, alpha=0)
        f0 = cost(x0)
        tau = max(0., min(1., tau))
        alpha = learning_rate
        x = self.update(params, grads, alpha)
        max_count = max(1, self.max_count)
        for _ in range(max_count):
            if cost(x) < f0:
                return x
            elif alpha < tol:
                return x
            else:
                alpha = learning_rate * tau
                x = self.update(params, grads, alpha)
                tau *= tau
        return x


class Optimizer:

    def __init__(self, line: LineSearch):
        self.line_search = line
        self.vars_history = None
        self.cost_history = None

    def minimize(self, x: List[np.ndarray], f: callable,
                 alpha: float = 0.01, max_iter: int = 100,
                 verbose: bool = False, epoch: int = None,
                 batch: int = None) -> List[np.ndarray]:
        """
        Minimize single objective function

        Parameters
        ----------
        x: List[np.ndarray]
            List of parameters to be updated

        f: callable
            Objective function y = f(x)

        alpha: float
            Learning rate
            Default = 0.01

        max_iter: int
            Maximum number of optimizer iterations allowed
            Default = 100

        verbose: bool
            Send progress output to standard out
            Default = False

        epoch: int
            The epoch in which this optimization is being run (only used
            for printing progress output messages)
            Default = None

        batch: int
            The batch in which this optimization is being run (only used
            for printing progress output messages)
            Default = None
        """
        line = self.line_search

        # Stopping criteria (Vanderplaats, ch. 3, p. 121)
        converged = False
        N1 = 0
        N1_max = 100
        N2 = 0
        N2_max = 100

        epsilon_absolute = 1e-6  # absolute error criterion
        epsilon_relative = 1e-6  # relative error criterion

        cost_history = []
        vars_history = []

        # Iterative update
        for i in range(0, max_iter):
            y = f(x)[0].squeeze()

            cost_history.append(y)
            vars_history.append(x)

            x = line.search(params=x, cost=lambda x: f(x)[0], grads=f(x)[1],
                            learning_rate=alpha)

            if verbose:
                if epoch is not None and batch is not None:
                    e = epoch
                    b = batch
                    print(f"epoch = {e:d}, batch = {b:d}, iter = {i:d}, cost = {y:6.3f}")
                elif epoch is not None:
                    e = epoch
                    print(f"epoch = {e:d}, iter = {i:d}, cost = {y:6.3f}")
                elif batch is not None:
                    b = batch
                    print(f"batch = {b:d}, iter = {i:d}, cost = {y:6.3f}")
                else:
                    print(f"iter = {i:d}, cost = {y:6.3f}")

            # Absolute convergence criterion
            if i > 1:
                dF1 = abs(cost_history[-1] - cost_history[-2])
                if dF1 < epsilon_absolute * cost_history[0]:
                    N1 += 1
                else:
                    N1 = 0
                if N1 > N1_max:
                    converged = True
                    if verbose:
                        print("Absolute stopping criterion satisfied")

                # Relative convergence criterion
                numerator = abs(cost_history[-1] - cost_history[-2])
                denominator = max(abs(cost_history[-1]), 1e-6)
                dF2 =  numerator / denominator

                if dF2 < epsilon_relative:
                    N2 += 1
                else:
                    N2 = 0
                if N2 > N2_max:
                    converged = True
                    if verbose:
                        print("Relative stopping criterion satisfied")

                if converged:
                    break

            # Maximum iteration convergence criterion
            if i == max_iter:
                if verbose:
                    print("Maximum optimizer iterations reached")

        self.cost_history = cost_history
        self.vars_history = vars_history

        return x


class GDOptimizer(Optimizer):

    def __init__(self):
        super().__init__(line=Backtracking(update=GD()))


class ADAMOptimizer(Optimizer):

    def __init__(self, beta_1: float = 0.9, beta_2: float = 0.99):
        adam = ADAM(beta_1, beta_2)
        super().__init__(line=Backtracking(update=adam))



from jenn._model import JENN
import numpy as np


def test_L_model_backward():
    """
    Fit a simple parabola twice: once using backprop and once again using
    finite difference. Make sure that the resulting coefficients are close.

    Note: it's important to fix the random seed for comparison
    """
    X = np.array([1, 2, 3, 4, 5]).reshape((-1, 1))
    Y = X ** 2
    J = (2 * X).reshape((-1, 1, 1))

    model = JENN(hidden_layer_sizes=(3,), is_finite_difference=False,
                 num_epochs=1, max_iter=10, alpha=0, gamma=0, random_state=0)

    model.fit(X.copy(), Y.copy(), J.copy())

    model_FD = JENN(hidden_layer_sizes=(3,), is_finite_difference=True,
                    num_epochs=1, max_iter=10, alpha=0, gamma=0, random_state=0)

    model_FD.fit(X.copy(), Y.copy(), J.copy())

    for i in range(len(model._W)):
        assert np.allclose(model._W[i], model_FD._W[i],
                           rtol=0.001)
        assert np.allclose(model._b[i], model_FD._b[i],
                           rtol=0.001)


if __name__ == "__main__":
    test_L_model_backward()



from jenn._loss import squared_loss, gradient_enhancement, regularization

import numpy as np


def test_squared_loss():

    # Check that perfect prediction returns 0.0 loss
    y_pred = np.array([1, 2, 3, 4, 5]).reshape((-1, 1))
    y_true = np.array([1, 2, 3, 4, 5]).reshape((-1, 1))
    assert np.allclose(squared_loss(y_true, y_pred), 0.)

    # Check that a prediction offset of exactly one
    # returns 1/2 times the number of examples times the number of responses
    y_pred = np.array([[1, 2, 3],
                       [4, 5, 6],
                       [7, 8, 9]]).reshape((-1, 1))
    y_true = np.array([[2, 3, 4],
                       [5, 6, 7],
                       [8, 9, 10]]).reshape((-1, 1))
    assert np.allclose(squared_loss(y_true, y_pred), 0.5 * 3 * 3)


def test_regularization():

    # Set coefficients equal to one and check
    # that, in that case, the answer is simply
    # the total number of coefficient times the
    # scale factor alpha / 2m
    alpha = 0.1
    m = 100
    n = 5
    w = [np.ones((n,))]
    actual = regularization(w, m, alpha)
    expected = len(w) * n * 0.5 * alpha / m
    assert actual == expected


def test_gradient_enhancement():

    # Check that perfect prediction of partials returns 0.0 loss
    dy_pred = np.array([[1, 2, 3],
                        [4, 5, 6],
                        [7, 8, 9]]).reshape((3, 3, 1))
    dy_true = np.array([[1, 2, 3],
                        [4, 5, 6],
                        [7, 8, 9]]).reshape((3, 3, 1))
    assert np.allclose(gradient_enhancement(dy_true, dy_pred), 0.)

    # Check that imperfect prediction of partials, with a constant
    # error of exactly 1 for each partial, returns the number of
    # examples (m=1) times the partials times the number of responses
    # scaled by gamma / 2m
    dy_pred = np.array([[1, 2, 3],
                        [4, 5, 6],
                        [7, 8, 9]]).reshape((3, 3, 1))
    dy_true = np.array([[2, 3, 4],
                        [5, 6, 7],
                        [8, 9, 10]]).reshape((3, 3, 1))
    assert np.allclose(gradient_enhancement(dy_true, dy_pred, gamma=1), 3 * 3 * 0.5)


if __name__ == "__main__":
    test_squared_loss()
    test_regularization()
    test_gradient_enhancement()



import numpy as np
from jenn._optimizer import Backtracking, ADAM, GD, ADAMOptimizer, GDOptimizer
from jenn.tests.test_problems import linear, rosenbrock, parabola
from importlib.util import find_spec

if find_spec("matplotlib"):
    from matplotlib import pyplot as plt
    MATPLOTLIB_INSTALLED = True
else:
    MATPLOTLIB_INSTALLED = False


def test_line_search():

    line = Backtracking(update=GD())

    f = lambda x: parabola(x)

    x0 = np.array([1]).reshape((1, 1))
    assert line.search([x0], f(x0)[1], f, learning_rate=0.1)[0] == 0.8

    x0 = np.array([-1]).reshape((1, 1))
    assert line.search([x0], f(x0)[1], f, learning_rate=2.0)[0] == 0

    f = lambda x: parabola(x, x0=1)

    x0 = np.array([1]).reshape((1, 1))
    assert line.search(x0, f(x0)[1], f, learning_rate=2.0)[0] == 1

    x0 = np.array([-1]).reshape((1, 1))
    assert line.search(x0, f(x0)[1], f, learning_rate=2.0)[0] == 1


def test_param_update():
    x0 = [np.array([5, 10]).reshape((1, -1))]
    y0, dydx = linear(x0)
    for update in [GD(), ADAM()]:
        x = update(x0, dydx, alpha=1)[0].squeeze()
        assert np.allclose(x, np.array([4, 9]))


def test_optimization(alpha: float = 0.05, max_iter: int = 1000,
                      is_adam: bool = True, is_plot: bool = False):
    """ check that optimizer yields correct answer for rosenbrock function"""

    # Initial guess
    x0 = [np.array([1.25, -1.75]).reshape((2, 1))]

    # Test function
    f = lambda x: rosenbrock(x)

    # Optimization
    if is_adam:
        opt = ADAMOptimizer()
    else:
        opt = GDOptimizer()

    xf = opt.minimize(x0, f, alpha=alpha, max_iter=max_iter)

    # For plotting contours
    lb = -2.
    ub = 2.
    m = 100
    x1 = np.linspace(lb, ub, m)
    x2 = np.linspace(lb, ub, m)
    X1, X2 = np.meshgrid(x1, x2)
    Y = np.zeros(X1.shape)
    for i in range(0, m):
        for j in range(0, m):
            X = [np.array([X1[i, j], X2[i, j]])]
            Y[i, j] = f(X)[0]

    if not MATPLOTLIB_INSTALLED:
        raise ImportError("Matplotlib must be installed.")

    if is_plot:
        x1_his = np.array([x[0][0] for x in opt.vars_history]).squeeze()
        x2_his = np.array([x[0][1] for x in opt.vars_history]).squeeze()
        plt.plot(x1_his, x2_his)
        plt.plot(x0[0][0], x0[0][1], '+', ms=15)
        plt.plot(xf[0][0], xf[0][1], 'o')
        plt.plot(np.array([1.]), np.array([1.]), 'x')
        plt.legend(
            ['history', 'initial guess', 'predicted optimum', 'true optimum'])
        plt.contour(X1, X2, Y, 50, cmap='RdGy')
        if is_adam:
            plt.title('ADAM')
        else:
            plt.title('GD')
        plt.show()

    # Close to optimum, the slope is nearly zero and the optimizer really
    # struggles to get to the exact optimum. +/- 0.2 is actually very close
    # to optimality. Turn on the contour plots to see.
    assert np.allclose(xf[0][0], 1.0, atol=0.2)
    assert np.allclose(xf[0][1], 1.0, atol=0.2)


if __name__ == "__main__":
    test_param_update()
    test_line_search()
    test_optimization(alpha=0.01, max_iter=10000, is_adam=True, is_plot=False)



"""
J A C O B I A N - E N H A N C E D   N E U R A L   N E T W O R K S  (J E N N)

Author: Steven H. Berguin <stevenberguin@gmail.com>

This package is distributed under the MIT license.
"""

import numpy as np
from typing import List


def regularization(w: List[np.ndarray], m: int, alpha: float = 0.) -> np.float:
    """
    Compute L2 norm penalty

    Parameters
    ----------
        w: List[np.ndarray]
            Weight parameters associated with each layer of neural net

        m: int
            Number of examples in training data

        alpha: float
            Regularization coefficient
    """
    alpha = max(0., alpha)  # ensure 0 < lambda
    penalty = 0.0
    for theta in w:
        penalty += np.squeeze(0.5 * alpha * np.sum(np.square(theta)))
    return 1. / m * np.float(penalty)


def gradient_enhancement(dy_true: np.ndarray, dy_pred: np.ndarray,
                         gamma: float = 1.) -> np.float:
    """
    Compute least squares estimator for the partials

    Parameters
    ----------
        dy_pred: np ndarray of shape (n_y, n_x, m)
            Predicted partials: AL' = d(AL)/dX where n_y = number outputs
                                                     n_x = number inputs
                                                     m = number examples

        dy_true: np ndarray of shape (n_y, n_x, m)
            True partials: Y' = d(Y)/dX

        gamma: float
            Regularization hyper-parameter
            Default = 1.0

    Returns
    -------
        loss: np.ndarray of shape (1,)
    """
    if gamma == 0:
        return 0
    n_y, n_x, m = dy_pred.shape  # number of outputs, inputs, training examples
    loss = 0.
    gamma = max(0., gamma)  # ensure 0 < gamma
    for k in range(0, n_y):
        for j in range(0, n_x):
            dy_j_pred = dy_pred[k, j, :].reshape(1, m)
            dy_j_true = dy_true[k, j, :].reshape(1, m)
            loss += np.squeeze(np.dot((dy_j_pred - dy_j_true),
                                      (dy_j_pred - dy_j_true).T))
    return 0.5 * gamma / m * np.float(loss)


def squared_loss(y_true: np.ndarray, y_pred: np.ndarray) -> np.float:
    """
    Compute least squares estimator for the states

    Parameters
    ----------
        y_pred: np.ndarray of shape (n_y, m)
            Predicted outputs where n_y = no. outputs, m = no. examples

        y_true: np.ndarray of shape (n_y, m)
            True outputs where n_y = no. outputs, m = no. examples
    """
    n_y, m = y_true.shape  # number of outputs, training examples
    cost = 0.
    for k in range(0, n_y):
        cost += np.squeeze(0.5 * np.dot((y_pred[k, :] - y_true[k, :]),
                                        (y_pred[k, :] - y_true[k, :]).T))

    return 1. / m * cost




from jenn._model import JENN
from jenn._utils import rsquare
import numpy as np

from importlib.util import find_spec

from jenn.tests.test_problems import rastrigin

if find_spec("matplotlib"):
    import matplotlib.pyplot as plt
    MATPLOTLIB_INSTALLED = True
else:
    MATPLOTLIB_INSTALLED = False

if find_spec("pyDOE2"):
    from pyDOE2 import lhs, fullfact
    PYDOE2_INSTALLED = True
else:
    PYDOE2_INSTALLED = False


def test_forward_prop():
    """
    Use a very simple network to check to check that
    forward propagation recovers expected results that
    can be computed by hand. Concretely, the following
    network is equivalent to Y = 4 * X
    """
    model = JENN(hidden_layer_sizes=(2, 2), activation='identity')
    hidden_activation = [model.activation] * len(model.hidden_layer_sizes)
    output_activation = ['identity']
    model._a = hidden_activation + output_activation
    model._W = [np.ones((2, 1)), np.ones((2, 2)), np.ones((1, 2))]
    model._b = [np.zeros((2, 1)), np.zeros((2, 1)), np.zeros((1, 1))]
    model._n_x = 1
    model._n_y = 1
    X = np.array([1, 2, 3, 4]).reshape((-1, 1))

    # States
    y_pred = model.predict(X).ravel()
    y_true = 4 * X.ravel()
    assert np.allclose(y_pred, y_true)

    # Partials
    dydx_pred = model.jacobian(X).ravel()
    dydx_true = np.array([4, 4, 4, 4])
    assert np.allclose(dydx_pred, dydx_true)


def test_parameter_shape():
    """
    Make sure that parameter initialization
    produces the correct parameter shapes
    """
    X = np.array([1, 2, 3, 4]).reshape((1, -1))
    y = np.array([1, 2, 3, 4]).reshape((1, -1))
    model = JENN(hidden_layer_sizes=(2, 2))
    model._n_x = X.shape[0]
    model._n_y = y.shape[0]
    model._initialize()

    assert model._W[0].shape == (2, 1)
    assert model._W[1].shape == (2, 2)
    assert model._W[2].shape == (1, 2)
    assert model._b[0].shape == (2, 1)
    assert model._b[1].shape == (2, 1)
    assert model._b[2].shape == (1, 1)


def test_model_parabola(verbose=False, show_plot=False):
    """
    Very simple test: fit a parabola. This test ensures that
    the model mechanics are working.
    """
    # Training data
    X = np.array([1, 2, 3, 4, 5]).reshape((-1, 1))
    Y = X ** 2
    J = (2*X).reshape((-1, 1, 1))

    # Basic neural net (no gradient-enhancement)
    model = JENN(hidden_layer_sizes=(3, 3), activation='tanh',
                 num_epochs=1, max_iter=1000,
                 learning_rate_init=0.05, alpha=0.01, gamma=0, verbose=verbose)
    model.fit(X, Y, J)
    if show_plot:
        model.goodness_fit(X, Y, title='NN')
    assert rsquare(Y, model.predict(X)) > 0.99

    # Gradient-enhanced neural net
    model = JENN(hidden_layer_sizes=(3, 3), activation='tanh',
                 num_epochs=1, max_iter=1000,
                 learning_rate_init=0.05, alpha=0.01, gamma=1, verbose=verbose)
    model.fit(X, Y, J)
    if show_plot:
        model.goodness_fit(X, Y, title='GENN')
    assert rsquare(Y, model.predict(X)) > 0.99


def test_model_rastrigin(verbose=False, show_plot=False):
    """ Test GENN on the rastrigin function (egg-crate looking function) """

    if not PYDOE2_INSTALLED:
        return None
    else:
        # Domain
        lb = -1.
        ub = 1.5

        # Training Data
        X_train = lb + (ub - lb) * lhs(2, samples=100, criterion='maximin',
                                       iterations=100, random_state=0)
        Y_train, J_train = rastrigin(X_train)

        # Test Data
        levels = 15
        X_test = fullfact([levels] * 2) / (levels - 1) * (ub - lb) + lb
        Y_test, J_test = rastrigin(X_test)

        # Train
        model = JENN(hidden_layer_sizes=[12] * 2, activation='tanh',
                     num_epochs=1, max_iter=1000,
                     is_finite_difference=False, solver='adam',
                     learning_rate='constant', random_state=0, tol=1e-6,
                     learning_rate_init=0.01, alpha=0, gamma=1,
                     verbose=verbose)

        model.fit(X_train, Y_train, J_train, is_normalize=True)

        if show_plot:
            model.goodness_fit(X_train, Y_train)
            model.goodness_fit(X_test, Y_test)

        assert rsquare(Y_train, model.predict(X_train)) > 0.95
        assert rsquare(Y_test, model.predict(X_test)) > 0.95


def test_sinusoid(verbose=False, show_plot=False, is_genn: bool = True):
    """
    Test GENN on simple sinusoid with very few points. Whereas regular NN
    will inevitably fail to provide a good fit, only GENN will succeed.
    """
    if show_plot and not MATPLOTLIB_INSTALLED:
        raise ImportError("Matplotlib must be installed.")

    # Test function
    f = lambda x: x * np.sin(x)

    # Test function derivative
    df_dx = lambda x: np.sin(x) + x * np.cos(x)

    # Domain
    lb = -np.pi
    ub = np.pi

    # Shapes
    n_x = 1  # number of inputs
    n_y = 1  # number of outputs

    # Training data
    m = 4  # number of training examples
    X_train = np.linspace(lb, ub, m).reshape((m, n_x))
    Y_train = f(X_train).reshape((m, n_y))
    J_train = df_dx(X_train).reshape((m, n_x, n_y))

    # Test data
    m = 30  # number of test examples
    X_test = lb + np.random.rand(m, 1).reshape((m, n_x)) * (ub - lb)
    Y_test = f(X_test).reshape((m, n_y))

    # Initialize model
    model = JENN(hidden_layer_sizes=(12,), activation='tanh',
                 num_epochs=1, max_iter=1000,
                 is_finite_difference=False,
                 learning_rate='backtracking', random_state=None, tol=1e-6,
                 learning_rate_init=0.05, alpha=0.1, gamma=int(is_genn),
                 verbose=verbose)

    # Train model
    model.fit(X_train, Y_train, J_train)

    # Predictions
    X = np.linspace(lb, ub, 100).reshape((100, n_x))
    Y_true = f(X)
    Y_pred = model.predict(X)

    if show_plot:

        if is_genn:
            title = 'GENN'
        else:
            title = 'NN'

        # Plot goodness of fit
        model.goodness_fit(X_test, Y_test)

        # Plot actual function
        fig, ax = plt.subplots()
        ax.plot(X, Y_pred, 'b-')
        ax.plot(X, Y_true, 'k--')
        ax.plot(X_test, Y_test, 'ro')
        ax.plot(X_train, Y_train, 'k+', mew=3, ms=10)
        ax.set(xlabel='x', ylabel='y', title=title)
        ax.legend(['Predicted', 'True', 'Test', 'Train'])
        plt.show()

    assert rsquare(Y_true, Y_pred) > 0.95


def run_tests():
    # test_forward_prop()
    # test_sinusoid(verbose=False, show_plot=False, is_genn=True)
    # test_parameter_shape()
    # test_model_parabola(verbose=False, show_plot=False)
    test_model_rastrigin(verbose=True, show_plot=False)


if __name__ == "__main__":
    run_tests()



import numpy as np
from typing import List, Tuple
from jenn._utils import finite_diff


def linear(x: List[np.ndarray], a: float = 1, b: float = 0,
           ) -> Tuple[np.ndarray, List[np.ndarray]]:
    x = x[0]
    y = a * np.sum(x) + b
    y = y.reshape(1, 1)
    dydx = np.array([a] * x.ndim).reshape(x.shape)
    return y, [dydx]


def parabola(x: List[np.ndarray], x0: float = 0
             ) -> Tuple[np.ndarray, List[np.ndarray]]:
    x = x[0]
    y = (x - x0) ** 2
    y = y.reshape(1, 1)
    dydx = (2 * (x - x0)).reshape(x.shape)
    return y, [dydx]


def rosenbrock(x: List[np.ndarray]) -> Tuple[np.ndarray, List[np.ndarray]]:
    x1 = x[0][0]
    x2 = x[0][1]
    y = (1 - x1) ** 2 + 100 * (x2 - x1 ** 2) ** 2
    y = y.reshape(1, 1)
    dydx = np.zeros(x[0].shape)
    dydx[0] = -2 * (1 - x1) - 400 * x1 * (x2 - x1 ** 2)
    dydx[1] = 200 * (x2 - x1 ** 2)
    return y, [dydx]


def rastrigin(X: np.ndarray) -> tuple:
    """
    Compute the Rastrigin function (and its gradient) given by:
                           n
    f(x1, ..., xn) = 10n + ∑ (xi ** 2 − 10 cos(2πxi))
                          i=1

    Parameters
    ----------
        X: List[np.ndarray]
            The inputs at which to evaluate the function
            shape = (m, n) where m = number of examples
                                 n = number of inputs (dimensionality)

    Returns
    -------
        y: np.ndarray
            The value of the response evaluated at X
            shape = (m, 1)

        dy_dX: np.ndarray
            The gradient of the response evaluated at X
            shape (m, n, 1)
    """
    m, n = X.shape
    y = np.zeros((m, 1)) + 10 * n
    dy_dX = np.zeros((m, n, 1))
    for i in range(n):
        x = X[:, i].reshape((-1, 1))
        y += np.power(x, 2) - 10 * np.cos(2 * np.pi * x)
        dy_dX[:, i] = 2 * x + 20 * np.pi * np.sin(2 * np.pi * x)
    return y, dy_dX


def test_rastrigin():
    """
    Check that Rastrigin returns correct values and partials
    by comparing against known answers at selected points.
    """

    x = np.zeros((1, 2))
    y, dy_dx = rastrigin(x)

    assert np.allclose(y.ravel(), 0.)
    assert np.allclose(dy_dx.ravel(),
                       finite_diff(x, f=lambda x: rastrigin(x)[0]))

    X = np.array([-0.259757925, 0.85149854]).reshape((1, 2))
    y, dy_dX = rastrigin(X)

    assert np.allclose(y.ravel(), np.array([15.45148376]))
    assert np.allclose(dy_dX.ravel(),
                       finite_diff(X, f=lambda x: rastrigin(x)[0]))


if __name__ == "__main__":
    test_rastrigin()


"""
J A C O B I A N - E N H A N C E D   N E U R A L   N E T W O R K S  (J E N N)

Author: Steven H. Berguin <stevenberguin@gmail.com>

This package is distributed under the MIT license.
"""

from ._activation import ACTIVATIONS

import numpy as np
from typing import List


def initialize_back_prop(AL: np.ndarray, Y: np.ndarray,
                         AL_prime: np.ndarray = None,
                         Y_prime: np.ndarray = None):
    """
    Initialize backward propagation

    Parameters
    ----------
        AL: np.ndarray
            predicted outputs
            shape = (n_y, m) where n_y = number outputs and m = number examples

        Y: np.ndarray
            true outputs values from training data
            shape = (n_y, m) where n_y = number outputs and m = number examples

        AL_prime: np.ndarray
            predicted Jacobian (i.e. predicted partials
                                w.r.t. inputs: AL' = d(AL)/dX)
            shape = (n_y, n_x, m) where n_x = number inputs,
                                        n_y = number outputs
                                        m = number examples
        
        Y_prime: np.ndarray
            actual Jacobian (i.e. true partials w.r.t. inputs: Y' = d(AL)/dX)
            shape = (n_y, n_x, m) where n_x = number inputs
                                        n_y = number outputs
                                        m = number examples

    Returns
    -------
        dAL: np.ndarray
            gradient of loss function w.r.t. last layer activations: d(L)/dAL
            shape = (n_y, m)

        dAL_prime: np.ndarray
            gradient of loss function w.r.t. last layer derivatives: d(L)/dAL'
            where AL' = d(AL)/dX
            shape = (n_y, n_x, m)
    """
    n_y, _ = AL.shape
    Y = Y.reshape(AL.shape)
    dAL = AL - Y
    dAL_prime = AL_prime - Y_prime
    return dAL, dAL_prime


def linear_activation_backward(dA: np.ndarray, dA_prime: np.ndarray,
                               cache: tuple, J_cache: tuple,
                               lambd: float, gamma: float) -> tuple:
    """
    Implement backward propagation for one LINEAR->ACTIVATION layer for the 
    regression least squares estimation

    Parameters
    ----------
    dA: np.ndarray
        gradient of loss function w.r.t. to post-activation
        for current layer l, dA = d(L)/dA where L is the loss function
        shape = (n_l, m) where n_l = number nodes in current layer
                               m = number of examples

    dA_prime: np.ndarray
        post-activation gradient w.r.t. A' for current layer l,
        dA' = d(L)/dA' where L is the loss function and A' = d(AL) / dX
        shape = shape (n_l, n_x, m) where n_l = number nodes in current layer
                                          n_x = number of inputs (X1, X2, ...)
                                          m = number of examples

    cache: Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, str]
        tuple = (A_prev, Z, W, b, activation)
            where
                A_prev -- activations from previous layer
                          shape = (n_prev, m) where n_prev is the number nodes in layer l-1
                Z -- input to activation functions for current layer
                    >> a numpy array of shape (n_x, m) where n_x is the number nodes in layer l
                W -- weight parameters for current layer
                    >> a numpy array of shape (n_x, n_prev)
                b -- bias parameters for current layer
                    >> a numpy array of shape (n_x, 1)
                activation -- activation function to use

    J_cache: np.ndarray
        list of caches containing precomputed values obtained during
        L_grads_forward() where J stands for Jacobian
        >> a list containing
            [..., (j, Z_prime_j, A_prime_j, G_prime, G_prime_prime), ...]
                   ------------------ input j --------------------
              where
                    j -- input variable associated with current cache
                              >> an integer representing the associated input
                                 variables (X1, X2, ..., Xj, ...)
                    Z_prime_j -- derivative of Z w.r.t. X_j: Z'_j = d(Z_j)/dX_j
                              >> a numpy array of shape (n_l, m) where n_l is
                                 the number nodes in layer l
                    A_prime_j -- derivative of the activation w.r.t. X_j
                                 A'_j = d(A_j)/dX_j
                              >> a numpy array of shape (n_l, m) where n_l
                                 is the number nodes in layer l

    lambd: float
        regularization parameter

    gamma: float
        gradient-enhancement parameter

    Returns
    -------
    dA_prev: np.ndarray
        gradient of the cost w.r.t. the activation (of the previous layer l-1)
        shape = same as A_prev = (n_prev, m)
        where n_prev is the number of nodes in layer l-1
              m the number of examples

    dW: np.ndarray
        gradient of the cost with respect to W (current layer l)
        shape = (n_x, n_prev) where n_prev is the number nodes in layer l-1
                                    n_x is the number of inputs

    db: np.ndarray
        gradient of the cost with respect to b (current layer l)
        shape = (n_x, 1) where n_x is the number of inputs
        
    dA_prime_prev: np.ndarray
        gradient of the cost w.r.t. activation partials of the previous layer
        shape = (n_prev, n_x, m) where  n_prev is the number nodes in layer l-1
                                        n_x is the number of inputs'
                                        m is the number of examples
    """
    # Extract information from current layer cache
    # A cache avoids recomputing what was already computed in fwd prop
    A_prev, Z, W, b, activation = cache

    # Activation function callable 
    g = ACTIVATIONS[activation]

    # Some dimensions that will be useful
    m = A_prev.shape[1]  # number of examples
    n_x = len(J_cache)  # number of inputs

    # 1st derivative of activation function A = G(Z)
    G_prime = g.first_derivative(Z)

    # Derivative of cost w.r.t. to parameters for this layer
    dW = 1. / m * np.dot(G_prime * dA, A_prev.T) + lambd / m * W
    db = 1. / m * np.sum(G_prime * dA, axis=1, keepdims=True)

    # Derivative of cost w.r.t. to activations of previous layer
    dA_prev = np.dot(W.T, G_prime * dA)

    # Initialize dA_prime_prev = d(J)/dA_prime_prev
    dA_prime_prev = np.zeros((W.shape[1], n_x, m))

    # Gradient enhancement
    if gamma != 0.0:

        # 2nd derivative of activation function A = G(Z)
        G_prime_prime = g.second_derivative(Z)

        # Loop over partials, d()/dX_j
        for j_cache in J_cache:
            # Extract information from current layer cache
            # associated with derivative of A w.r.t. j^th input
            j, Z_prime_j, A_prime_j_prev = j_cache

            # Extract partials of A w.r.t. to j^th input
            # i.e. A_prime_j = d(A)/dX_j
            dA_prime_j = dA_prime[:, j, :].reshape(Z_prime_j.shape)

            # Compute contribution to cost function gradient
            # db = d(J)/db, dW = d(J)/dW, d(L)/dA, d(L)/dA'
            dW += gamma / m * (np.dot(dA_prime_j * G_prime_prime * Z_prime_j,
                                      A_prev.T) +
                               np.dot(dA_prime_j * G_prime, A_prime_j_prev.T))
            db += gamma / m * np.sum(dA_prime_j * G_prime_prime * Z_prime_j,
                                     axis=1, keepdims=True)
            dA_prev += gamma * np.dot(W.T,
                                      dA_prime_j * G_prime_prime * Z_prime_j)
            dA_prime_prev[:, j, :] = gamma * np.dot(W.T, dA_prime_j * G_prime)

    return dA_prev, dW, db, dA_prime_prev


def L_model_backward(AL: np.ndarray, Y: np.ndarray, caches: list,
                     AL_prime: np.ndarray = None, Y_prime: np.ndarray = None,
                     J_caches: list = None,
                     lambd: float = 0, gamma: float = 0) -> tuple:
    """
    Implement backward propagation for the entire neural network 

    Parameters
    ----------
    AL: np.ndarray
        predicted outputs
        shape = (n_y, m) where n_y = number outputs and m = number examples

    Y: np.ndarray
        true outputs values from training data
        shape = (n_y, m) where n_y = number outputs and m = number examples

    AL_prime: np.ndarray
        predicted Jacobian (i.e. partials w.r.t. inputs: AL' = d(AL)/dX)
        shape = (n_y, n_x, m) where n_x = number inputs
                                    n_y = number outputs
                                    m = number examples

    Y_prime: np.ndarray
        actual Jacobian (i.e. true partials w.r.t. inputs: Y' = d(AL)/dX)
        shape = (n_y, n_x, m)

    caches: List[np.ndarray]
        list of tuples, where each tuple contains the precomputed values
        stored during fwd_prop.linear_activation_forward() for each layer
        tuple = (A_prev, Z, W, b, activation)
            where
                A_prev -- activations from previous layer
                          shape = (n_prev, m)
                          where n_prev is the number nodes in layer l-1
                Z -- input to activation functions for current layer
                    >> a numpy array of shape (n_x, m)
                       where n_x is the number nodes in layer l
                W -- weight parameters for current layer
                    >> a numpy array of shape (n_x, n_prev)
                b -- bias parameters for current layer
                    >> a numpy array of shape (n_x, 1)
                activation -- activation function to use

    J_caches: List[np.ndarray]
        list of caches containing precomputed values obtained during L_grads_forward() where J stands for Jacobian
              >> a tuple [ [[...], ..., [...]], ..., [..., (j, Z_prime_j, A_prime_j, G_prime, G_prime_prime), ...], ...]
                            --- layer 1 ------        ------------------ layer l, partial j ---------------------
                      where
                            j -- input variable number (i.e. X1, X2, ...) associated with cache
                                      >> an integer representing the associated input variables (X1, X2, ..., Xj, ...)
                            Z_prime_j -- derivative of Z w.r.t. X_j: Z'_j = d(Z_j)/dX_j
                                      >> a numpy array of shape (n_l, m) where n_l is the number nodes in layer l
                            A_prime_j -- derivative of the activation w.r.t. X_j: A'_j = d(A_j)/dX_j
                                      >> a numpy array of shape (n_l, m) where n_l is the number nodes in layer l

    alpha: float
        regularization hyper-parameter

    gamma: float
        gradient-enhancement hyper-parameter

    Returns
    -------
    dW: [np.ndarray]
        Gradient of cost w.r.t. weights for each layer

    db: [np.ndarray]
        Gradient of cost w.r.t. intercepts for each layer
    """
    dWs = []
    dbs = []

    # Some quantities needed
    L = len(caches)  # the number of layers
    _, m = AL.shape
    Y = Y.reshape(AL.shape)  # after this line, Y is the same shape as AL

    # Initializing the back propagation
    dA, dA_prime = initialize_back_prop(AL, Y, AL_prime, Y_prime)

    # Loop over each layer
    for l in reversed(range(L)):
        # Get cache
        cache = caches[l]
        J_cache = J_caches[l]

        # Backprop step
        dA, dW, db, dA_prime = linear_activation_backward(dA, dA_prime, cache,
                                                          J_cache, lambd,
                                                          gamma)

        # Store result
        dWs.append(dW)
        dbs.append(db)

    # Gradients are needed in the original order
    dWs = list(reversed(dWs))
    dbs = list(reversed(dbs))

    return dWs, dbs



"""
J A C O B I A N - E N H A N C E D   N E U R A L   N E T W O R K S  (J E N N)

Author: Steven H. Berguin <stevenberguin@gmail.com>

This package is distributed under the MIT license.
"""

import numpy as np
from collections import defaultdict
from typing import Tuple, Union, List
from ._fwd_prop import L_model_forward, L_grads_forward
from ._bwd_prop import L_model_backward
from ._utils import (mini_batches, grad_check, finite_diff, goodness_of_fit,
                     DataConverter, scale_factor)
from ._loss import squared_loss, gradient_enhancement, regularization
from ._optimizer import Optimizer, Backtracking, ADAM, GD

from importlib.util import find_spec

if find_spec("matplotlib"):
    import matplotlib.pyplot as plt
    MATPLOTLIB_INSTALLED = True
else:
    MATPLOTLIB_INSTALLED = False

EPS = np.finfo(float).eps  # small number to avoid division by zero


class GENNBase:
    """ Gradient-Enhanced Neural Net (GENN) """

    def __init__(self,
                 hidden_layer_sizes: Union[List[int], tuple] = (100,),
                 activation: str = "relu", *,
                 solver: str = 'adam',
                 alpha: float = 0.0001,
                 gamma: int = 1,
                 batch_size: int = None,
                 learning_rate: str = "constant",
                 learning_rate_init: float = 0.001,
                 num_epochs: int = 1,
                 max_iter: int = 100,
                 shuffle: bool = True,
                 random_state: int = None,
                 tol: float = 1e-4,
                 verbose: bool = False,
                 warm_start: bool = False,
                 beta_1: float = 0.9,
                 beta_2: float = 0.999,
                 is_finite_difference: bool = False,
                 is_grad_check: bool = False):
        """
        Initialize object

        Parameters
        ----------
        hidden_layer_sizes: List[int]
            The ith element represents the number of neurons
            in the ith hidden layer
            Default = (110,)

        activation:str
            Activation function for the hidden layer.
            option = {‘identity’, ‘logistic’, ‘tanh’, ‘relu’}
            Default = ’tanh’

        solver: str
            The type of solver to use for gradient-based optimization
            Options = {‘sgd’, ‘adam’}
            Default = 'adam'

        alpha: float
            L2 penalty (regularization term) parameter
            Default = 1e-4

        gamma:
            Parameter controlling influence of partials
            squared loss term during training
            Default = 1

        batch_size: str or int
            Size of mini-batches for stochastic optimizers.
            When set to “auto”, batch_size=min(200, n_samples)
            Default = 'auto'

        learning_rate: str
            Learning rate schedule for weight updates.
            Options =
                ‘constant’ = constant learning rate given by learning_rate_init
                ‘backtracking’ = learning is optimized during each line search
            Default = 'constant'

        learning_rate_init: float
            The initial learning rate used. It controls the step-size
            during line search, when updating the neural net parameters.
            Default = 1e-3

        num_epochs: int
            Number of passes through data
            Default = 1

        max_iter: int
            Number of optimizer iterations. Note that optimizer is called once
            per batch for each epoch. For example: 10 epochs, 5 batches, 100
            max_iter results in 10 * 5 * 100 = 5_000 function calls to train.
            Default = 100

        shuffle: bool
            Whether to shuffle batch samples for each epoch
            Default = True

        random_state: int
            Fix the random state to ensure repeatable results
            Default = None

        tol: float
            Optimizer tolerance
            Default = 1e-4

        verbose: bool
            Whether to print progress messages to stdout
            Default = False

        warm_start: bool
            When set to True, reuse the solution of the previous call to fit
            as initialization, otherwise, just erase the previous solution
            Default = False

        beta_1: float
            Exponential decay rate for estimates of first moment vector
            in adam, should be in [0, 1). Only used when solver=’adam’
            Default = 0.9

        beta_2: float
            Exponential decay rate for estimates of second moment vector
            in adam, should be in [0, 1). Only used when solver=’adam’
            Default = 0.99

        is_finite_difference: bool
            This option is intended for debugging. Finite difference is
            used instead of backpropagation during training.
            Default = False

        is_grad_check: bool
            This option is intended for debugging. The analytical gradient
            computed using back-prop is checked against finite difference at
            every optimizer iteration.
            Default = False

        normalize: bool
            Normalize the training data according to mean and variance. This
            can often greatly help optimization as it makes the optimization
            problem easier to navigate and solve efficiently.
            Default = False
        """

        self.activation = activation
        self.solver = solver
        self.alpha = alpha
        self.gamma = gamma
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.learning_rate_init = learning_rate_init
        self.num_epochs = num_epochs
        self.max_iter = max_iter
        self.hidden_layer_sizes = list(hidden_layer_sizes)
        self.shuffle = shuffle
        self.random_state = random_state
        self.tol = tol
        self.verbose = verbose
        self.warm_start = warm_start
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.is_finite_difference = is_finite_difference
        self.is_grad_check = is_grad_check

        # Dimensionality of inputs and outputs
        self._n_x = None
        self._n_y = None

        # Parameters
        self._W = None  # coefficients
        self._b = None  # intercepts
        self._a = None  # activations

        # Training history
        self._cost_history = None

    def _cost(self, W: List[np.ndarray], b: List[np.ndarray],
              X: np.ndarray, Y: np.ndarray, J: np.ndarray = None,
              is_grad: bool = True):
        """

        Parameters
        ----------
        W: List[np.ndarray]
        b: List[np.ndarray]
        X: np.ndarray
        Y: np.ndarray
        J: np.ndarray
        is_grad: bool

        Return
        ------
        c, dW, db : Tuple[np.float, List[np.ndarray], List[np.ndarray]]
            c = cost
            dW = d(cost)/dW = derivative of cost w.r.t. neural net weights
            db = d(cost)/db = derivative of cost w.r.t. neural net biases
        """
        # Predict
        Y_pred, caches = L_model_forward(X, W, b, self._a)
        J_pred, J_caches = L_grads_forward(X, W, b, self._a)

        # Cost function
        c = 0
        c = c + squared_loss(Y, Y_pred)
        c = c + regularization(W, X.shape[1], self.alpha)
        if J is not None:
            c = c + gradient_enhancement(J_pred, J, self.gamma)

        # Cost function gradient
        dW = []
        db = []
        if is_grad:
            dW, db = L_model_backward(Y_pred, Y, caches, J_pred, J, J_caches,
                                      self.alpha, self.gamma)
        return c, dW, db

    def _initialize(self):
        """ Initialize neural net using "He" initialization """
        if not self._n_y:
            raise ValueError(f'Missing output layer: n_y = {self._n_y}')

        if not self.hidden_layer_sizes:
            raise ValueError(
                "Neural net does not seem to have any hidden layers")

        np.random.seed(self.random_state)

        layer_dims = [self._n_x] + self.hidden_layer_sizes + [self._n_y]

        # Parameters
        self._W = []
        self._b = []
        number_layers = len(layer_dims)
        for l in range(1, number_layers):
            W = np.random.randn(layer_dims[l], layer_dims[l - 1]) \
                * np.sqrt(1. / layer_dims[l - 1])
            b = np.zeros((layer_dims[l], 1))
            self._W.append(W)
            self._b.append(b)

    def _train(self, X: np.ndarray, Y: np.ndarray, J: np.ndarray = None):
        """
        Internal support function in charge of training the neural net

        Parameters 
        ----------
        X: matrix of shape (n_x, m) where n_x = no. of inputs
                                          m = no. of training examples
        Y: matrix of shape (n_y, m) where n_y = no. of outputs
        J: np.ndarray of size (n_y, n_x, m) where J = Jacobian
                                                      dY1/dX1 = J[0, 0, :]
                                                      dY1/dX2 = J[0, 1, :]
                                                      ...
                                                      dY2/dX1 = J[1, 0, :]
                                                      dY2/dX2 = J[1, 1, :]
                                                      ...

        Returns
        -------
        cost_history : defaultdict(dict)
            Cost function values for each epoch and batch:
                cost_history["epoch_e"]["batch_b"] = c
                where e is the epoch number
                      b is the batch number
                      c is the cost as an np.ndarray (c[i] = cost of i^th iter)
        """
        if X.ndim != 2:
            raise ValueError(f'X.ndim = {X.ndim} but X.shape should (n_x, m)')

        if Y.ndim != 2:
            raise ValueError(f'Y.ndim = {Y.ndim} but X.shape should (n_y, m)')

        self._n_x = X.shape[0]
        self._n_y = Y.shape[0]

        # Initialize
        if not self.warm_start or self._W is None or self._b is None:
            self._initialize()
        else:
            # Check that user did not change dimensions
            if len(self._W) - 1 != len(self.hidden_layer_sizes):
                raise ValueError(
                    f'warm start flag raised, but number of hidden layers '
                    f'no longer matches previously trained neural net.')
            if self._n_x != X.shape[0]:
                raise ValueError(
                    f'warm start flag raised, but new input layer does '
                    f'not match shape of previous input layer.')
            if self._n_y != Y.shape[0]:
                raise ValueError(
                    f'warm start flag raised, but new output layer does '
                    f'not match shape of previous output layer.')

        hidden_activation = [self.activation] * len(self.hidden_layer_sizes)
        output_activation = ['identity']
        self._a = hidden_activation + output_activation

        cost_history = defaultdict(dict)

        n_x, m = X.shape
        n_y, _ = Y.shape

        if J is None:
            J = np.zeros((n_y, n_x, m))
            self.gamma = 0  # Make sure gradient-enhancement is off if J = None

        for e in range(self.num_epochs):
            batches = mini_batches(X, self.batch_size,
                                   random_state=self.random_state,
                                   shuffle=self.shuffle)
            for b, batch in enumerate(batches):

                X_batch = X[:, batch]
                Y_batch = Y[:, batch]
                J_batch = J[:, :, batch]

                def f(x, is_grad=True):
                    """ Cost function and gradient """
                    W = x[:len(self._W)]
                    b = x[len(self._W):]
                    y, dW, db = self._cost(W, b, X_batch, Y_batch, J_batch,
                                           is_grad)
                    dy_dx = dW + db
                    return y, dy_dx

                def f_FD(x):
                    """ Debug: cost function with finite difference grad """
                    y = f(x, is_grad=False)[0]
                    dy_dx = finite_diff(x, lambda x: f(x, is_grad=False)[0])
                    return y, dy_dx

                def f_check(x):
                    """ Debug: cost function with FD vs. analytic grad """
                    grad_check(x, f=lambda x: f(x, is_grad=False)[0],
                               dfdx=lambda x: f(x, is_grad=True)[1])
                    if self.is_finite_difference:
                        return f_FD(x)
                    else:
                        return f(x)

                cost = f
                if self.is_grad_check:
                    cost = f_check
                elif self.is_finite_difference:
                    cost = f_FD

                if self.solver == 'adam':
                    update = ADAM(self.beta_1, self.beta_2)
                elif self.solver == 'sgd':
                    update = GD()
                else:
                    raise ValueError(f'solver = {self.solver} not recognized')

                if self.learning_rate == 'constant':
                    line_search = Backtracking(update,
                                               max_count=0, tol=self.tol)
                elif self.learning_rate == 'backtracking':
                    line_search = Backtracking(update, tol=self.tol)
                else:
                    raise ValueError(f'learning_rate = '
                                     f'{self.learning_rate} not recognized')

                optimizer = Optimizer(line_search)

                params = optimizer.minimize(x=self._W + self._b, f=cost,
                                            alpha=self.learning_rate_init,
                                            max_iter=self.max_iter,
                                            verbose=self.verbose,
                                            epoch=e, batch=b)

                self._W = params[:len(self._W)]
                self._b = params[len(self._W):]

                key1 = 'epoch_' + str(e)
                key2 = 'batch_' + str(b)
                cost_history[key1][key2] = np.array(optimizer.cost_history)

        self._cost_history = cost_history


class JENN(GENNBase):

    def __init__(self, *args, **kwargs):
        self._data_converter = None
        super().__init__(*args, **kwargs)

    def fit(self, X: np.ndarray, Y: np.ndarray, J: np.ndarray = None,
            is_normalize: bool = False):
        """
        Minimize augmented squared loss in order to learn neural net parameters

        Parameters
        ----------
        X: np.ndarray
            Training data features
            shape = (m, n_x) where m = no. examples, n_x = no. features

        Y: np.ndarray
            Training data responses
            shape = (m, n_y) where n_y = no. responses

        J: np.ndarray
            Training data Jacobian, J = dY/dX
            shape = (m, n_x, n_y)

        is_normalize: bool
            Option to normalize data by mean and variance before fitting
            Default = False
        """
        X = X.T  # The algorithm assumes X.shape = (n_x, m)
        Y = Y.T  # The algorithm assumes Y.shape = (n_y, m)
        if J is not None:
            J = J.T  # The algorithm assumes J.shape = (n_y, n_x, m)

        if is_normalize:
            mu_x, sigma_x = scale_factor(X)
            mu_y, sigma_y = scale_factor(Y)
            self._data_converter = DataConverter(mu_x, sigma_x, mu_y, sigma_y)
        else:
            self._data_converter = None

        if self._data_converter is not None:
            X_norm = self._data_converter.X_norm(X)
            Y_norm = self._data_converter.Y_norm(Y)
            J_norm = self._data_converter.J_norm(J)
            self._train(X_norm, Y_norm, J_norm)
        else:
            self._train(X, Y, J)

    def predict(self, X: np.ndarray):
        """
        Predict response(s) using trained neural network

        Parameters
        ----------
            X: np.ndarray
                Training data features
                shape = (m, n_x) where m = no. examples, n_x = no. features

        Returns
        -------
            Y: np.ndarray
                Training data responses
                shape = (m, n_y) where n_y = no. responses
        """
        X = X.T  # The algorithm assumes X.shape = (n_x, m)
        if self._data_converter is not None:
            X_norm = self._data_converter.X_norm(X)
            Y_norm = L_model_forward(X_norm, self._W, self._b, self._a,
                                     store_cache=False)
            Y = self._data_converter.Y(Y_norm)
        else:
            Y = L_model_forward(X, self._W, self._b, self._a,
                                store_cache=False)
        return Y.T  # The algorithm returns Y.shape = (n_y, m)

    def jacobian(self, X: np.ndarray):
        """
        Predict Jacobian using trained neural network

        Parameters
        ----------
        X: np.ndarray
            Training data features
            shape = (m, n_x) where m = no. examples, n_x = no. features

        Returns
        -------
        J: np.ndarray
            Training data Jacobian, J = dY/dX of shape = (m, n_x, n_y)
        """
        X = X.T  # The algorithm assumes X.shape = (n_x, m)
        if self._data_converter is not None:
            X_norm = self._data_converter.X_norm(X)
            J_norm = L_grads_forward(X_norm, self._W, self._b, self._a,
                                     store_cache=False)
            J = self._data_converter.J(J_norm)
        else:
            J = L_grads_forward(X, self._W, self._b, self._a,
                                store_cache=False)
        return J.T  # The algorithm returns J.shape = (n_y, n_x, m)

    def goodness_fit(self, X: np.ndarray, Y_true: np.ndarray,
                     title: str = None, legend: str = None,
                     show_error: bool = True):
        """
        Plot goodness of fit

        Parameters
        ----------
        X: np.ndarray
            Inputs, shape = (m, n_x) where m = no. examples, n_x = no. features

        Y_true: np.ndarray
            Outputs, shape = (m, n_y) where m = no. examples, n_y = no. outputs
        """
        if not MATPLOTLIB_INSTALLED:
            raise ImportError("Matplotlib must be installed.")

        if self._n_x != X.shape[1]:
            raise ValueError(f'X.shape[1] = {X.shape[1]} '
                             f'but expected {self._n_x}')

        if self._n_y != Y_true.shape[1]:
            raise ValueError(f'Y_true.shape[1] = {Y_true.shape[1]} '
                             f'but expected {self._n_y}')

        Y_pred = self.predict(X)

        figs = []
        for i in range(self._n_y):
            y_pred = Y_pred[:, i].ravel()
            y_true = Y_true[:, i].ravel()
            if title is None:
                title = f'Goodness of Fit: Y[{i}]'
            fig = goodness_of_fit(y_pred, y_true, title, legend, show_error)
            figs.append(fig)

        return figs

    def training_history(self, title: str = 'Training History'):

        if not MATPLOTLIB_INSTALLED:
            raise ImportError("Matplotlib must be installed.")

        if not self.training_history:
            return None

        epochs = list(self._cost_history.keys())
        if len(epochs) > 1:
            avg_costs = []
            for epoch in epochs:
                batches = self._cost_history[epoch].keys()
                avg_batch_costs = []
                for batch in batches:
                    avg_batch_cost = self._cost_history[epoch][batch].mean()
                    avg_batch_costs.append(avg_batch_cost)
                avg_cost = sum(avg_batch_costs) / len(batches)
                avg_costs.append(avg_cost)
            plt.plot(range(len(epochs)), avg_costs)
            plt.xlabel('epoch')
            plt.ylabel('avg cost')
        elif len(self._cost_history['epoch_0']) > 1:
            avg_cost = []
            batches = self._cost_history['epoch_0'].keys()
            for batch in batches:
                avg_cost.append(self._cost_history['epoch_0'][batch].mean())
            plt.plot(range(len(batches)), avg_cost)
            plt.xlabel('batch')
            plt.ylabel('avg cost')
        else:
            cost = self._cost_history['epoch_0']['batch_0']
            plt.plot(np.arange(cost.size), cost)
            plt.xlabel('iteration')
            plt.ylabel('cost')
        plt.title(title)
